//
// © 2005-2008 Marko Lindqvist
//

#include "sdksui.h"

#include <iostream>

#include "cell.h"
#include "sdksoption.h"
#include "sudoku.h"
#include "sdksinput.h"

sdks_ui::~sdks_ui()
{}

void sdks_ui::solve_log(cell *pcell, const char *msg)
{
  if(get_option(OPTION_EXPLAIN)) {
    std::cout << "(" << pcell->get_x() << "," << pcell->get_y() << "):";
    std::cout << msg << std::endl;
  }
}

void sdks_ui::start_output()
{}

void sdks_ui::finish_output()
{}

void sdks_ui::print_header(std::string msg, int size)
{
  print_msg(msg);
}

void sdks_ui::print_msg(std::string msg)
{
  std::cout << msg << std::endl;
}

void sdks_ui::print_sudoku(sudoku *puzzle)
{
  int ry, rx, y, x, ri, i;
  int size = puzzle->get_size();

  for(ry = 0; ry < size; ry++) {
    for (ri = 0; ri < size; ri++) {
      std::cout << "+";
      for (i = 0; i < size; i++) {
        std::cout << "-";
      }
    }
    std::cout << "+\n";
    for(y = 0; y < size; y++) {
      for(rx = 0; rx < size; rx++) {
        std::cout << "|";
        for(x = 0; x < size; x++) {
          puzzle->get_cell(CT_ROW, ry*size+y, rx*size+x)->print("?", "!", this, size);
        }
      }
      std::cout << "|\n";
    }
  }
  for (ri = 0; ri < size; ri++) {
    std::cout << "+";
    for (i = 0; i < size; i++) {
      std::cout << "-";
    }
  }
  std::cout << "+\n";
}


void sdks_ui::print_sudoku_possibilities(sudoku *puzzle)
{
  int ry, rx, y, x, i, j, vy, vx;
  int count = 0;
  int size = puzzle->get_size();
  cell *pcell;

  // Print header
  std::cout << "  ";
  for (rx = 0; rx < size; rx++) {
    std::cout << "  ";
    for (i = 0; i < size; i++) {
      for (j = 0; j < (size / 2); j++) {
        std::cout << " ";
      }
      std::cout << int_to_char(count++, size);
      for (j++; j < size; j++) {
        std::cout << " ";
      }
      std::cout << " "; // Space between cells
    }
  }

  std::cout << "\n";
  count = 0; // Y-coordinate now on

  for (ry = 0; ry < size; ry++) {
    print_possibility_ruler(size);
    for (y = 0; y < size; y++) {
      print_possibility_verticals(size);
      for (vy = 0; vy < size; vy++) {
        if (vy == size / 2) {
          // Line starts with Y-coordinate
          std::cout << int_to_char(count++, size) << " ";
        } else {
          std::cout << "  ";
        }

        for (rx = 0; rx < size; rx++) {
          std::cout << "| ";
          for (x = 0; x < size; x++) {
            pcell = puzzle->get_cell(CT_ROW, ry*size+y, rx*size+x);
            if (pcell->solved()) {
              if (vy == (size - 1) / 2) {
                for (i = 0; i < size / 2; i++) {
                  std::cout << "*";
                }
                std::cout << int_to_char(pcell->solved(), size);
                i++;
                for (;i < size; i++) {
                  std::cout << "*";
                }
              } else {
                for (i = 0; i < size; i++) {
                  std::cout << "-";
                }
              }
            } else {
              for (vx = 0; vx < size; vx++) {
                if (pcell->is_possible(vy*size+vx+1)) {
                  std::cout << int_to_char(vy*size+vx+1, size);
                } else {
                  std::cout << "-";
                }
              }
            }
            std::cout << " "; // Space between cells
          }
        }
        std::cout << "|\n";
      }
    }
    print_possibility_verticals(size);
  }
  print_possibility_ruler(size);
}


void sdks_ui::print_possibility_ruler(int size)
{
  int ri, i;

  std::cout << "  +";
  for (ri = 0; ri < size; ri++) {
    std::cout << "=";
    for (i = 0; i < size*(size+1); i++) {
      // Each cell needs base_size characters + space between cells
      std::cout << "=";
    }
    std::cout << "+";
  }
  std::cout << "\n";
}


void sdks_ui::print_possibility_verticals(int size)
{
  int ri, i;

  std::cout << "  |";
  for (ri = 0; ri < size; ri++) {
    std::cout << " ";
    for (i = 0; i < size*(size+1); i++) {
      // Each cell needs size characters + space between cells
      std::cout << " ";
    }
    std::cout << "|";
  }
  std::cout << "\n";
}


void sdks_ui::print_method_name(const char *mname)
{
  std::cout << mname << std::endl;
}


char sdks_ui::int_to_char(int n, int size)
{
  if (n < 10) {
    return '0' + n;
  }

  if (size > 3 && n == size*size) {
    // Hex sudokus (and other big sudokus) have '0' as high end character
    return '0';
  }

  if (n < 25) {
    return 'A' + n - 10;
  }

  return '?';
}


int sdks_ui::char_to_int(char c, int size)
{
  if (c >= '1' && c <= '9') {
    return c - '0';
  }
  if (c >= 'a' && c <= 'n') {
    return c + 10 - 'a';
  }
  if (c >= 'A' && c <= 'N') {
    return c + 10 - 'A';
  }
  if (size > 3 && c == '0') {
    // High end character
    return size*size;
  }

  return -1;
}

sdks_input *sdks_ui::get_input(int origbase)
{
  int dimension = origbase*origbase;
  int i, j;
  int srcj = 0;
  std::string line;
  sdks_input *input = new sdks_input(origbase);

  for(i = 0; i < dimension; i++) {
    bool line_found = false;

    do {
      getline(std::cin, line);
      if (line.length() > 0) {
	line_found = true;
	srcj = 0;

	for (j = 0; j < dimension; srcj++) {
	  int value;

	  if (line[srcj] != ' ') {
	    if (line[srcj] == '\0') {
	      std::cerr << "Line " << i + 1 << " too short!" << std::endl;
	      input->set_fail();
	      return input;
	    }
	    value = char_to_int(line[srcj], origbase);

	    if (line[srcj] != '?' &&
		(value < 1 || value > origbase*origbase)) {
	      std::cerr << "Illegal character '" << line[srcj] << "' in line ";
	      std::cerr << i + 1 << "." << std::endl;
	      input->set_fail();
	      return input;
	    }

	    if (line[srcj] == '?') {
	      input->set_value(j, i, -1);
	    } else {
	      input->set_value(j, i, line[srcj] - '0');
	    }
	    j++;
	  }
	}
	if (line[srcj] != '\0') {
	  std::cerr << "Line " << i + 1 << " too long!" << std::endl;
	  input->set_fail();
	  return input;
	}
      }
    } while(!line_found);
  }

  return input;
}
